/********************************************
 *  pmacV3Controller.h
 *
 * PMAC V3 controller driver based on the asynMotorController class
 *
 * Stefan Mathis, September 2024
 ********************************************/

#ifndef pmacV3Controller_H
#define pmacV3Controller_H
#include "pmacV3Axis.h"
#include "sinqAxis.h"
#include "sinqController.h"

#define IncrementalEncoder "Incremental encoder"
#define AbsoluteEncoder "Absolute encoder"

class pmacV3Controller : public sinqController {

  public:
    /**
     * @brief Construct a new pmacV3Controller object
     *
     * @param portName            See sinqController constructor
     * @param ipPortConfigName    See sinqController constructor
     * @param numAxes             See sinqController constructor
     * @param movingPollPeriod    See sinqController constructor
     * @param idlePollPeriod      See sinqController constructor
     * @param comTimeout          When trying to communicate with the device,
    the underlying asynOctetSyncIO interface waits for a response until this
    time (in seconds) has passed, then it declares a timeout.
     */
    pmacV3Controller(const char *portName, const char *ipPortConfigName,
                     int numAxes, double movingPollPeriod,
                     double idlePollPeriod, double comTimeout);

    /**
     * @brief Get the axis object
     *
     * @param pasynUser           Specify the axis via the asynUser
     * @return pmacV3Axis*        If no axis could be found, this is a nullptr
     */
    pmacV3Axis *getAxis(asynUser *pasynUser);

    /**
     * @brief Get the axis object
     *
     * @param axisNo              Specify the axis via its index
     * @return pmacV3Axis*        If no axis could be found, this is a nullptr
     */
    pmacV3Axis *getAxis(int axisNo);

    /**
     * @brief Overloaded function of asynMotorController
     *
     * The function is overloaded to allow enabling / disabling the motor and
     * rereading the encoder.
     *
     * @param pasynUser           Specify the axis via the asynUser
     * @param value               New value
     * @return asynStatus
     */
    asynStatus writeInt32(asynUser *pasynUser, epicsInt32 value);

    /**
     * @brief Overloaded function of asynMotorController
     *
     * The function is overloaded to get readback values for the enabling /
     * disabling status and the encoder.
     *
     * @param pasynUser           Specify the axis via the asynUser
     * @param value               Read-out value
     * @return asynStatus
     */
    asynStatus readInt32(asynUser *pasynUser, epicsInt32 *value);

  protected:
    asynUser *lowLevelPortUser_;

    /**
     * @brief Send a command to the hardware and receive a response
     *
     * @param axisNo              Axis to which the command should be send
     * @param command             Command for the hardware
     * @param response            Buffer for the response. This buffer is
     * expected to have the size MAXBUF_.
     * @param numExpectedResponses  The PMAC MCU can send multiple responses at
     * once. The number of responses is determined by the number of
     * "subcommands" within command. Therefore it is known in advance how many
     * "subresponses" are expected. This can be used to check the integrity of
     * the received response, since the subresponses are separated by carriage
     * returns (/r). The number of carriage returns is compared to
     * numExpectedResponses to determine if the communication was successfull.
     * @return asynStatus
     */
    asynStatus writeRead(int axisNo, const char *command, char *response,
                         int numExpectedResponses);

    /**
     * @brief Save cast of the given asynAxis pointer to a pmacV3Axis pointer.
     * If the cast fails, this function returns a nullptr.
     *
     * @param asynAxis
     * @return pmacV3Axis*
     */
    pmacV3Axis *castToAxis(asynMotorAxis *asynAxis);

  private:
    // Set the maximum buffer size. This is an empirical value which must be
    // large enough to avoid overflows for all commands to the device /
    // responses from it.
    static const uint32_t MAXBUF_ = 200;

    /*
    Stores the constructor input comTimeout
    */
    double comTimeout_;

    // Indices of additional PVs
    int enableMotor_;
    int motorEnabled_;
    int rereadEncoderPosition_;
    int rereadEncoderPositionRBV_;
    int readConfig_;
    int encoderType_;

    /*
    These parameters are here to write the high and low limits from the MCU to
    the EPICS motor record. Using motorHighLimit_ / motorLowLimit_ does not
    work: https://epics.anl.gov/tech-talk/2023/msg00576.php.
    Therefore, some additional records are introduced which read from these
    parameters and write into the motor record. See the sinq_asyn_motor.db file.
    */
    int motorHighLimitFromDriver_;
    int motorLowLimitFromDriver_;

    friend class pmacV3Axis;
};

#endif /* pmacV3Controller_H */
